// primitives.js
//   This code uses the libraries from the WebGL book,

//----------- declare globals  ------------------------------------------------------

var gl;              // GL rendering context
var u_FragColor;     // 4x1 color vector on GPU
var u_MvpMatrix;     // 4x4 modeling transformation matrix on GPU
var mvpMatrix = new Matrix4();  // create new 4x4 matrix
var g_time = 0;      // global elapsed time, as measured in seconds
var canvas;          // canvas that WebGL will use to draw

//----------- main()   --------------------------------------------------------------

function main() {
  var foo;

  console.log('hello world');

  createPrimitives();       // create the object geometry: face, cube, pyramid, sphere, cylinder
  canvas = mySetup();       // setup shaders, send vertices to the GPU

    // Set clear color and enable hidden surface removal
  gl.clearColor(0.9, 0.95, 0.96, 1.0);
  gl.enable(gl.DEPTH_TEST);

    // Register the event handler to be called on key press
  document.onkeydown = function(ev){ keydown(ev); };

  animate();
//  draw();
}

//----------- animate()   --------------------------------------------------------------

function animate() {
  var tick = function() {                // create the callback function
    update();
    draw();
    requestAnimationFrame(tick, canvas); // Request that the browser calls tick
  };
  tick();                // now call it once. Further calls will be via the requests
}

//----------- update()   --------------------------------------------------------------

var g_startTime = Date.now();   // last time the function was called, in milliseconds

function update() {
  g_time = (Date.now() - g_startTime)/1000;
}

//----------- draw()   --------------------------------------------------------------

var th0=0;

function draw() {
  var white = [1,1,1];
  var green = [0.4,0.5,0.4];
  var purple = [0.6, 0.6, 0.8];
  var brown = [0.7, 0.5, 0.1];
  var lightBrown = [0.8,0.6,0.3];

  // Clear color and depth buffer
  gl.clear(gl.COLOR_BUFFER_BIT | gl.DEPTH_BUFFER_BIT);

  // Set the eye point and the viewing volume
  mvpMatrix.setPerspective(30, 1, 1, 100);
  mvpMatrix.lookAt(8,16,30, 0,5,0,  0,1,0);

  mvpMatrix.rotate(th0, 0,1,0);
  drawPrimitives();

    // draw body
  mvpMatrix.rotate(90, 0,0,1);
  drawBlock(4,2,1, green);

  pushMatrix(mvpMatrix);    // save a copy of F_body

    // draw head
  mvpMatrix.translate(4.5,0,0);
  drawBlock(1.5,1.5,1, white);

  mvpMatrix = popMatrix();  // return to body frame
  
    // draw upper arm
  mvpMatrix.translate(3,0,0.3);
  mvpMatrix.rotate(-60, 0,0,1);
  mvpMatrix.rotate(30*Math.sin(g_time), 0,0,1);     // rotate the model

  drawBlock(3,1,1, brown);

    // draw lower arm
  mvpMatrix.translate(3,0,0);
  mvpMatrix.rotate(20, 0,0,1);
  mvpMatrix.rotate(30*Math.sin(g_time), 0,0,1);     // rotate the model
  drawBlock(3,1,1, lightBrown);

    // draw ball
  setColor(purple[0], purple[1], purple[2]);
  mvpMatrix.translate(4,0,0);
  drawPrimitive("sphere");
}

//----------- drawPrimitives()   --------------------------------------------------------------

function drawPrimitives() {
  pushMatrix(mvpMatrix);    
  
  setColor(1,1,1);
  mvpMatrix.translate(-2,-2,0);
  drawPrimitive("cube");

  setColor(1,1,1);
  mvpMatrix.translate(2,0,0);
  drawPrimitive("face");

  setColor(1,1,1);
  mvpMatrix.translate(2,0,0);
  drawPrimitive("cylinder");

  setColor(1,1,1);
  mvpMatrix.translate(2,0,0);
  drawPrimitive("pyramid");

  setColor(1,1,1);
  mvpMatrix.translate(2,0,0);
  drawPrimitive("sphere");

  mvpMatrix = popMatrix();
}

//----------- drawBlock()   --------------------------------------------------------------
// draws a block of the given dimensions, and that spans the following ranges:
//  x = [0,xSize]     y = [-0.5*ySize, ySize]    z = [-0.5*zSize, zSize]

function drawBlock(xSize, ySize, zSize, colour) {
  pushMatrix(mvpMatrix);

   // scale original 2x2x2 cube to be the desired size
  mvpMatrix.scale(0.5*xSize,0.5*ySize,0.5*zSize); 
   // original cube is centered; translate so that the final cube has
  mvpMatrix.translate(1,0,0);     // l
  gl.uniformMatrix4fv(u_MvpMatrix, false, mvpMatrix.elements);  // send to GPU

    // Draw the cube
  setColor(colour[0],colour[1],colour[2]);
  drawPrimitive("cube");

  mvpMatrix = popMatrix();
}

//----------- pushMatrix()   -----------------------------------------------------------

var g_matrixStack = []; // Array for storing a matrix
function pushMatrix(m) { // Store the specified matrix to the array
  var m2 = new Matrix4(m);
  g_matrixStack.push(m2);
}

//----------- popMatrix()   -----------------------------------------------------------

function popMatrix() { // Retrieve the matrix from the array
  return g_matrixStack.pop();
}

//----------- setColor()   -----------------------------------------------------------

function setColor(r,g,b) {
    alpha = 1.0;
    gl.uniform4f(u_FragColor,r,g,b,alpha);   // send color to GPU
}

//----------- keydown()   -----------------------------------------------------------

function keydown(ev) {
    if(ev.keyCode == 65) { // 'a' key pressed
      console.log("pressed a key");
      th0 -= 10;
    } else 
    if (ev.keyCode == 83) { // 's' key pressed
      console.log("pressed A key");
      th0 += 10;
    } else { return; }
//    draw();
}

//----------- mySetup()   -----------------------------------------------------------

var VSHADER_SOURCE =
  'attribute vec4 a_Position;\n' +
  'uniform mat4 u_MvpMatrix;\n' +
  'void main() {\n' +
  '  gl_Position = u_MvpMatrix * a_Position;\n' +
  '}\n';

// Fragment shader program
var FSHADER_SOURCE =
  '#ifdef GL_ES\n' +
  'precision mediump float;\n' +
  '#endif\n' +
  'uniform vec4 u_FragColor;\n' + 
  'void main() {\n' +
  '  gl_FragColor = u_FragColor;\n' +
  '}\n';

function mySetup() {
  var canvas = document.getElementById('webgl');
  gl = getWebGLContext(canvas);          // get rendering context 
  if (!gl) throw 'Failed to get the rendering context for WebGL';

  // Initialize shaders
  if (!initShaders(gl, VSHADER_SOURCE, FSHADER_SOURCE)) {
    console.log('Failed to intialize shaders.');
    throw ' ';
  }
 
  buildBuffers();       // save vertex + index info to buffers, send to GPU

 // Get the storage location of u_MvpMatrix
  u_MvpMatrix = gl.getUniformLocation(gl.program, 'u_MvpMatrix');
  if (!u_MvpMatrix) { 
    console.log('Failed to get the storage location of u_MvpMatrix');
    return;
  }

  // Get location of color vector 
  u_FragColor = gl.getUniformLocation(gl.program, 'u_FragColor');
  if (!u_FragColor) throw 'Failed to get storage locn of u_FragColor';

  return canvas;
}

//----------- declare primitives and their cumulateive vertex + index lists -----------------

function createPrimitives() {
    // procedural creation of objects
  createCylinder("cylinder",10);
  createSphere("sphere",6,12);

    // create objects from a vertices list & indices list
    // vertices list:  provide x,y,z triples
    // indices list:   provide  i1,i2,i3,i4  for every desired quadrilateral,
    //                 where i=0 refers to the first vertex,
    //                       i=1 refers to the second vertex, etc.

  createGeometry("face",[0,0,0,    1,0,0,   1,1,0,  0,1,0], [0,1,2,3]);
  createGeometry("pyramid",[-1,-1,0,    1,-1,0,   1,1,0,  -1,1,0,  0,0,1], 
     [0,1,2,3,   0,1,4,4,  1,2,4,4,  2,3,4,4,  3,0,4,4 ]);
  createGeometry("cube",[     // vertices
     1.0,  1.0,  1.0, // v0 
    -1.0,  1.0,  1.0, // v1 
    -1.0, -1.0,  1.0, // v2 
     1.0, -1.0,  1.0, // v3 
     1.0, -1.0, -1.0, // v4 
     1.0,  1.0, -1.0, // v5 
    -1.0,  1.0, -1.0, // v6 
    -1.0, -1.0, -1.0, // v7 
   ], [       // indices 
    0, 1, 2, 3,   // front:  v0,v1,v2,v3
    0, 3, 4, 5,   // right:  v0,v3,v4,v5
    0, 5, 6, 1,   // up:     etc.
    1, 6, 7, 2,   // left
    7, 4, 3, 2,   // down 
    4, 7, 6, 5    // back
    ]);
}

var vertexList = [];
var indexList = [];
var primitives = [];

//----------- createGeometry() ----------------------------------------------------

function createGeometry(objName, vertices, indices) {
  var baseIndx = indexList.length;
  var baseVert = vertexList.length;
  for (var n=0; n<vertices.length; n+=3) 
      vertexList.push(vertices[n+0], vertices[n+1], vertices[n+2]);
  for (var n=0; n<indices.length; n++)
      indexList.push(baseVert/3 + indices[n]);
  primitives.push({ name:objName, startIndx:baseIndx, faceCount:indices.length/4});
  console.log('Created primitive ',objName, 'nVertices = ',vertices.length/3, 'nFaces = ',indices.length/4);
}

//----------- cylinder ----------------------------------------------------

function createCylinder(cylName,nSlices) {
  var vertices = [];
  var indices = [];
  var slice;
  var dth = 2*Math.PI/nSlices;
  for (slice=0; slice<=nSlices; slice++) {
    var sin_th = Math.sin(slice*dth);
    var cos_th = Math.cos(slice*dth);
    vertices.push(cos_th,sin_th,0);
    vertices.push(cos_th,sin_th,1);
  }
  for (slice=0; slice<=nSlices-1; slice++) {
    i = slice*2;
    indices.push(i+0);
    indices.push(i+1);
    indices.push(i+3);
    indices.push(i+2);
  }
  createGeometry(cylName, vertices, indices);
}

//----------- sphere ----------------------------------------------------

function createSphere(sphereName,nLatSlices, nLongSlices) {
  var slice;
  var baseIndx = indexList.length;
  var baseVertex = vertexList.length/3;
  var dth_lat = Math.PI/nLatSlices;
  var dth_long = 2*Math.PI/nLongSlices;
  for (var i=0; i<=nLatSlices; i++) {
    var th_lat = -0.5*Math.PI + i*dth_lat;
    var sin_lat = Math.sin(th_lat);
    var cos_lat = Math.cos(th_lat);
    for (var j=0; j<=nLongSlices; j++) {
      var th_long = j*dth_long;
      var sin_long = Math.sin(th_long);
      var cos_long = Math.cos(th_long);
      var r = cos_lat;
      var z = sin_lat;
      var x = r*cos_long;
      var y = r*sin_long;
      vertexList.push(x,y,z);
    }
  }
  for (var i=0; i<nLatSlices; i++) {
    for (var j=0; j<nLongSlices; j++) {
      var indx = baseVertex + i*(nLongSlices+1) + j;
      indexList.push(indx);
      indexList.push(indx+1);
      indexList.push(indx+1+(nLongSlices+1));
      indexList.push(indx+(nLongSlices+1));
    }
  }

  var nVert = vertexList.length/3 - baseVertex;
  var nFaces = (indexList.length - baseIndx)/4;
  primitives.push({ name:sphereName, startIndx:baseIndx, faceCount:nFaces});
  console.log('Created primitive ',sphereName, 'startIndx = ',baseIndx, 'faceCount = ',nFaces, 
     'nVert = ',nVert);
}

// ---------------- buildBuffers ------------------------

function buildBuffers() {
//  console.log('buildBuffers()');

  var nCoords;
  nCoords = vertexList.length;
  var vertexF32List = new Float32Array(nCoords);
  for (var n=0; n<nCoords; n++) 
    vertexF32List[n] = vertexList[n];

  nCoords = indexList.length;
  var indexU16List = new Uint16Array(nCoords);
  for (var n=0; n<nCoords; n++) 
    indexU16List[n] = indexList[n];

  var vertexBuffer = gl.createBuffer();
  var indexBuffer = gl.createBuffer();
  if (!vertexBuffer || !indexBuffer)   return -1;

  // Write the vertex coordinates to the buffer object
  gl.bindBuffer(gl.ARRAY_BUFFER, vertexBuffer);   // set as current array buffer
  gl.bufferData(gl.ARRAY_BUFFER, vertexF32List, gl.STATIC_DRAW);  // allocate & copy data

  var FSIZE = vertexF32List.BYTES_PER_ELEMENT;

    // Assign the buffer object to a_Position and enable the assignment
  var a_Position = gl.getAttribLocation(gl.program, 'a_Position');
  if(a_Position < 0) {
    console.log('Failed to get the storage location of a_Position');
    return -1;
  }
    // specify where to find the a_position attribute
  gl.vertexAttribPointer(a_Position, 3, gl.FLOAT, false, FSIZE * 3, 0);  
  gl.enableVertexAttribArray(a_Position);

  // Write the indices to the buffer object
  gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, indexBuffer);
  gl.bufferData(gl.ELEMENT_ARRAY_BUFFER, indexU16List, gl.STATIC_DRAW);

  return indexU16List.length;
}

// ------------------ drawPrimitive() ----------------------------

function drawPrimitive(pname) {
  var nFaces=0;
  for (var n=0; n < primitives.length; n++) {
    if (primitives[n].name === pname) {
      pbase = primitives[n].startIndx;
      nFaces = primitives[n].faceCount;
//      console.log('found, pbase = ', pbase);
    }
  } 

//  console.log('drawPrimitive');
//  console.log('nFaces = ', nFaces);

  if (nFaces==0) throw 'drawPrimitive: primitive not found';

  gl.uniformMatrix4fv(u_MvpMatrix, false, mvpMatrix.elements);  // send to GPU

    // Draw the faces
  for (n=0; n<nFaces; n++) 
    gl.drawElements(gl.TRIANGLE_FAN, 4, gl.UNSIGNED_SHORT, 2*(pbase + n*4));

    // Draw the edges
  setColor(0,0,0);
  for (n=0; n<nFaces; n++) 
    gl.drawElements(gl.LINE_LOOP, 4, gl.UNSIGNED_SHORT, 2*(pbase + n*4));
}
